# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import logging
import re
from asyncio import iscoroutinefunction
from functools import partial
from inspect import Signature, isasyncgenfunction, _empty
from typing import TYPE_CHECKING, Any, Callable, TypeVar, Union, List

from . import __package__
from .chat_command_handler import ChatCommandHandler, _hash_app_command_params
from ..commands.arg_types import (
    ChannelTypes,
    CommandArg,
    Description,
    Choices,
    MaxValue,
    MinValue,
)
from ..commands.groups import Group, Subgroup
from ..utils.snowflake import Snowflake
from ..utils.types import APINullable, MISSING
from ..exceptions import (
    CommandIsNotCoroutine,
    CommandAlreadyRegistered,
    TooManyArguments,
    InvalidArgumentAnnotation,
    CommandDescriptionTooLong,
    InvalidCommandGuild,
    InvalidCommandName,
)
from ..objects import (
    ThrottleScope,
    AppCommand,
    Role,
    User,
    Channel,
    Mentionable,
    MessageContext,
)
from ..objects.app import (
    AppCommandOptionType,
    AppCommandOption,
    InteractableStructure,
    AppCommandType,
)
from ..utils import should_pass_ctx
from ..utils.signature import get_signature_and_params

if TYPE_CHECKING:
    from typing import Optional

REGULAR_COMMAND_NAME_REGEX = re.compile(r"[\w\- ]{1,32}$")
CHAT_INPUT_COMMAND_NAME_REGEX = re.compile(r"^[a-z0-9_-]{1,32}$")

_log = logging.getLogger(__package__)

_options_type_link = {
    Signature.empty: AppCommandOptionType.STRING,
    str: AppCommandOptionType.STRING,
    int: AppCommandOptionType.INTEGER,
    bool: AppCommandOptionType.BOOLEAN,
    float: AppCommandOptionType.NUMBER,
    User: AppCommandOptionType.USER,
    Channel: AppCommandOptionType.CHANNEL,
    Role: AppCommandOptionType.ROLE,
    Mentionable: AppCommandOptionType.MENTIONABLE,
}

if TYPE_CHECKING:
    from ..client import Client

T = TypeVar("T")


def command(
    func=None,
    *,
    name: Optional[str] = None,
    description: Optional[str] = "Description not set",
    enable_default: Optional[bool] = True,
    guild: Union[Snowflake, int, str] = None,
    cooldown: Optional[int] = 0,
    cooldown_scale: Optional[float] = 60.0,
    cooldown_scope: Optional[ThrottleScope] = ThrottleScope.USER,
    parent: Optional[Union[Group, Subgroup]] = None,
):
    """A decorator to create a slash command to register and respond to
    with the discord API from a function.

    str - String
    int - Integer
    bool - Boolean
    float - Number
    pincer.objects.User - User
    pincer.objects.Channel - Channel
    pincer.objects.Role - Role
    pincer.objects.Mentionable - Mentionable

    .. code-block:: python3

        class Bot(Client):
            @command(
                name="test",
                description="placeholder"
            )
            async def test_command(
                self,
                ctx: MessageContext,
                amount: int,
                name: CommandArg[
                    str,
                    Description("Do something cool"),
                    Choices(Choice("first value", 1), 5)
                ],
                optional_int: CommandArg[
                    int,
                    MinValue(10),
                    MaxValue(100),
                ] = 50
            ):
                return Message(
                    f"You chose {amount}, {name}, {letter}",
                    flags=InteractionFlags.EPHEMERAL
                )


    References from above:
        :class:`~pincer.client.Client`,
        :class:`~pincer.objects.message.message.Message`,
        :class:`~pincer.objects.message.context.MessageContext`,
        :class:`~pincer.objects.app.interaction_flags.InteractionFlags`,
        :class:`~pincer.commands.arg_types.Choices`,
        :class:`~pincer.commands.arg_types.Choice`,
        :class:`typing_extensions.Annotated` (Python 3.8),
        :class:`typing.Annotated` (Python 3.9+),
        :class:`~pincer.commands.arg_types.CommandArg`,
        :class:`~pincer.commands.arg_types.Description`,
        :class:`~pincer.commands.arg_types.MinValue`,
        :class:`~pincer.commands.arg_types.MaxValue`


    Parameters
    ----------
    name : Optional[:class:`str`]
        The name of the command |default| :data:`None`
    description : Optional[:class:`str`]
        The description of the command |default| ``Description not set``
    enable_default : Optional[:class:`bool`]
        Whether the command is enabled by default |default| :data:`True`
    guild : Optional[Union[:class:`~pincer.utils.snowflake.Snowflake`, :class:`int`, :class:`str`]]
        What guild to add it to (don't specify for global) |default| :data:`None`
    cooldown : Optional[:class:`int`]
        The amount of times in the cooldown_scale the command can be invoked
        |default| ``0``
    cooldown_scale : Optional[:class:`float`]
        The 'checking time' of the cooldown |default| ``60``
    cooldown_scope : :class:`~pincer.objects.app.throttle_scope.ThrottleScope`
        What type of cooldown strategy to use |default| :attr:`ThrottleScope.USER`

    Raises
    ------
    CommandIsNotCoroutine
        If the command function is not a coro
    InvalidCommandName
        If the command name does not follow the regex ``^[\\w-]{1,32}$``
    InvalidCommandGuild
        If the guild id is invalid
    CommandDescriptionTooLong
        Descriptions max 100 characters
        If the annotation on an argument is too long (also max 100)
    CommandAlreadyRegistered
        If the command already exists
    TooManyArguments
        Max 25 arguments to pass for commands
    InvalidArgumentAnnotation
        Annotation amount is max 25,
        Not a valid argument type,
        Annotations must consist of name and value
    """  # noqa: E501
    if func is None:
        return partial(
            command,
            name=name,
            description=description,
            enable_default=enable_default,
            guild=guild,
            cooldown=cooldown,
            cooldown_scale=cooldown_scale,
            cooldown_scope=cooldown_scope,
            parent=parent,
        )

    cmd = name or func.__name__

    if not re.match(CHAT_INPUT_COMMAND_NAME_REGEX, cmd):
        raise InvalidCommandName(
            f"Command `{cmd}` doesn't follow the name requirements."
            " Ensure to match the following regex:"
            f" {CHAT_INPUT_COMMAND_NAME_REGEX.pattern}"
        )

    options: List[AppCommandOption] = []

    signature, params = get_signature_and_params(func)
    pass_context = should_pass_ctx(signature, params)

    if len(params) > (25 + pass_context):
        cmd = name or func.__name__
        raise TooManyArguments(
            f"Command `{cmd}` (`{func.__name__}`) can only have 25 "
            f"arguments (excluding the context and self) yet {len(params)} "
            "were provided!"
        )

    for idx, param in enumerate(params):
        if idx == 0 and pass_context:
            continue

        sig = signature[param]

        annotation, required = sig.annotation, sig.default is _empty

        # ctx is type MessageContext but should not be included in the
        # slash command
        if annotation == MessageContext and idx == 1:
            return

        argument_type = None
        if type(annotation) is CommandArg:
            argument_type = annotation.command_type
        # isinstance and type don't work for Annotated. This is the best way 💀
        elif hasattr(annotation, "__metadata__"):
            # typing.get_origin doesn't work in 3.9+ for some reason. Maybe they forgor
            # to implement it.
            argument_type = annotation.__origin__

        # check if None to keep the type checker happy that its Any
        if argument_type is not None and hasattr(argument_type, "__args__"):
            # this is a Union, hopefully an Optional/Union[T, None]
            # Optional[T] is an alias to Union[T, None]
            args = argument_type.__args__

            if len(args) != 2 or args[1] != type(None):
                raise InvalidArgumentAnnotation(
                    "`Union` is not a supported option type"
                )

            argument_type = args[0]

        if not argument_type:
            if annotation in _options_type_link:
                options.append(
                    AppCommandOption(
                        type=_options_type_link[annotation],
                        name=param,
                        description="Description not set",
                        required=required,
                    )
                )
                continue

            # TODO: Write better exception
            raise InvalidArgumentAnnotation(
                "Type must be Annotated or other valid type"
            )

        command_type = _options_type_link[argument_type]

        def get_arg(t: T) -> APINullable[T]:
            if type(annotation) is CommandArg:
                return annotation.get_arg(t)
            elif hasattr(annotation, "__metadata__"):
                for obj in annotation.__metadata__:
                    if isinstance(obj, t):
                        return obj.get_payload()
                return MISSING

        argument_description = get_arg(Description) or "Description not set"

        choices = get_arg(Choices)

        if choices is not MISSING and argument_type not in {
            int,
            float,
            str,
        }:
            raise InvalidArgumentAnnotation(
                "Choice type is only allowed for str, int, and float"
            )
        if choices is not MISSING:
            for choice in choices:
                if isinstance(choice.value, int) and argument_type is float:
                    continue
                if not isinstance(choice.value, argument_type):
                    raise InvalidArgumentAnnotation(
                        "Choice value must match the command type"
                    )

        channel_types = get_arg(ChannelTypes)
        if channel_types is not MISSING and argument_type is not Channel:
            raise InvalidArgumentAnnotation(
                "ChannelTypes are only available for Channels"
            )

        max_value = get_arg(MaxValue)
        min_value = get_arg(MinValue)

        for i, value in enumerate((min_value, max_value)):
            if (
                value is not MISSING
                and argument_type is not int
                and argument_type is not float
            ):
                t = ("MinValue", "MaxValue")
                raise InvalidArgumentAnnotation(
                    f"{t[i]} is only available for int and float"
                )

        options.append(
            AppCommandOption(
                type=command_type,
                name=param,
                description=argument_description,
                required=required,
                choices=choices,
                channel_types=channel_types,
                max_value=max_value,
                min_value=min_value,
            )
        )

    # Discord API returns MISSING for options when there are 0. Options is set MISSING
    # so equality checks later work properly.
    if not options:
        options = MISSING

    return register_command(
        func=func,
        app_command_type=AppCommandType.CHAT_INPUT,
        name=name,
        description=description,
        enable_default=enable_default,
        guild=guild,
        cooldown=cooldown,
        cooldown_scale=cooldown_scale,
        cooldown_scope=cooldown_scope,
        command_options=options,
        parent=parent,
    )


def user_command(
    func=None,
    *,
    name: Optional[str] = None,
    enable_default: Optional[bool] = True,
    guild: Union[Snowflake, int, str] = None,
    cooldown: Optional[int] = 0,
    cooldown_scale: Optional[float] = 60,
    cooldown_scope: Optional[ThrottleScope] = ThrottleScope.USER,
):
    """A decorator to create a user command registering and responding
    to the Discord API from a function.

    .. code-block:: python3

         class Bot(Client):
             @user_command
             async def test_user_command(
                 self,
                 ctx: MessageContext,
                 user: User,
                 member: GuildMember
             ):
                 if not member:
                     # member is missing if this is a DM
                     # This bot doesn't like being DMed, so it won't respond
                     return

                 return f"Hello {user.name}, this is a Guild."


    References from above:
        :class:`~client.Client`,
        :class:`~objects.message.context.MessageContext`,
        :class:`~objects.user.user.User`,
        :class:`~objects.guild.member.GuildMember`,


    Parameters
    ----------
    name : Optional[:class:`str`]
        The name of the command |default| :data:`None`
    enable_default : Optional[:class:`bool`]
        Whether the command is enabled by default |default| :data:`True`
    guild : Optional[Union[:class:`~pincer.utils.snowflake.Snowflake`, :class:`int`, :class:`str`]]
        What guild to add it to (don't specify for global) |default| :data:`None`
    cooldown : Optional[:class:`int`]
        The amount of times in the cooldown_scale the command can be invoked
        |default| ``0``
    cooldown_scale : Optional[:class:`float`]
        The 'checking time' of the cooldown |default| ``60``
    cooldown_scope : :class:`~pincer.objects.app.throttle_scope.ThrottleScope`
        What type of cooldown strategy to use |default| :attr:`ThrottleScope.USER`

    Raises
    ------
    CommandIsNotCoroutine
        If the command function is not a coro
    InvalidCommandName
        If the command name does not follow the regex ``^[\\w-]{1,32}$``
    InvalidCommandGuild
        If the guild id is invalid
    CommandDescriptionTooLong
        Descriptions max 100 characters
        If the annotation on an argument is too long (also max 100)
    CommandAlreadyRegistered
        If the command already exists
    InvalidArgumentAnnotation
        Annotation amount is max 25,
        Not a valid argument type,
        Annotations must consist of name and value
    """  # noqa: E501
    if func is None:
        return partial(
            user_command,
            name=name,
            enable_default=enable_default,
            guild=guild,
            cooldown=cooldown,
            cooldown_scale=cooldown_scale,
            cooldown_scope=cooldown_scope,
        )

    return register_command(
        func=func,
        app_command_type=AppCommandType.USER,
        name=name,
        enable_default=enable_default,
        guild=guild,
        cooldown=cooldown,
        cooldown_scale=cooldown_scale,
        cooldown_scope=cooldown_scope,
    )


def message_command(
    func=None,
    *,
    name: Optional[str] = None,
    enable_default: Optional[bool] = True,
    guild: Union[Snowflake, int, str] = None,
    cooldown: Optional[int] = 0,
    cooldown_scale: Optional[float] = 60,
    cooldown_scope: Optional[ThrottleScope] = ThrottleScope.USER,
):
    """A decorator to create a user command to register and respond
    to the Discord API from a function.

    .. code-block:: python3

        class Bot(Client):
            @user_command
            async def test_message_command(
                self,
                ctx: MessageContext,
                message: UserMessage,
            ):
                return message.content


    References from above:
        :class:`~client.Client`,
        :class:`~objects.message.context.MessageContext`,
        :class:`~objects.message.message.UserMessage`,
        :class:`~objects.user.user.User`,
        :class:`~objects.guild.member.GuildMember`,


    Parameters
    ----------
    name : Optional[:class:`str`]
        The name of the command |default| :data:`None`
    enable_default : Optional[:class:`bool`]
        Whether the command is enabled by default |default| :data:`True`
    guild : Optional[Union[:class:`~pincer.utils.snowflake.Snowflake`, :class:`int`, :class:`str`]]
        What guild to add it to (don't specify for global) |default| :data:`None`
    cooldown : Optional[:class:`int`]
        The amount of times in the cooldown_scale the command can be invoked
        |default| ``0``
    cooldown_scale : Optional[:class:`float`]
        The 'checking time' of the cooldown |default| ``60``
    cooldown_scope : :class:`~pincer.objects.app.throttle_scope.ThrottleScope`
        What type of cooldown strategy to use |default| :attr:`ThrottleScope.USER`

    Raises
    ------
    CommandIsNotCoroutine
        If the command function is not a coro
    InvalidCommandName
        If the command name does not follow the regex ``^[\\w-]{1,32}$``
    InvalidCommandGuild
        If the guild id is invalid
    CommandDescriptionTooLong
        Descriptions max 100 characters
        If the annotation on an argument is too long (also max 100)
    CommandAlreadyRegistered
        If the command already exists
    InvalidArgumentAnnotation
        Annotation amount is max 25,
        Not a valid argument type,
        Annotations must consist of name and value
    """  # noqa: E501
    if func is None:
        return partial(
            message_command,
            name=name,
            enable_default=enable_default,
            guild=guild,
            cooldown=cooldown,
            cooldown_scale=cooldown_scale,
            cooldown_scope=cooldown_scope,
        )

    return register_command(
        func=func,
        app_command_type=AppCommandType.MESSAGE,
        name=name,
        enable_default=enable_default,
        guild=guild,
        cooldown=cooldown,
        cooldown_scale=cooldown_scale,
        cooldown_scope=cooldown_scope,
    )


def register_command(
    func: Callable[..., Any] = None,
    app_command_type: Optional[AppCommandType] = None,
    name: Optional[str] = None,
    description: Optional[str] = MISSING,
    enable_default: Optional[bool] = True,
    guild: Optional[Union[Snowflake, int, str]] = None,
    cooldown: Optional[int] = 0,
    cooldown_scale: Optional[float] = 60.0,
    cooldown_scope: Optional[ThrottleScope] = ThrottleScope.USER,
    command_options=MISSING,  # Missing typehint?
    parent: Optional[Union[Group, Subgroup]] = MISSING,
):
    cmd = name or func.__name__

    if not re.match(REGULAR_COMMAND_NAME_REGEX, cmd):
        raise InvalidCommandName(
            f"Command `{cmd}` doesn't follow the name requirements."
            " Ensure to match the following regex:"
            f" {REGULAR_COMMAND_NAME_REGEX.pattern}"
        )

    if not iscoroutinefunction(func) and not isasyncgenfunction(func):
        raise CommandIsNotCoroutine(
            f"Command with call `{func.__name__}` is not a coroutine, "
            "which is required for commands."
        )

    try:
        guild_id = Snowflake(guild) if guild else MISSING
    except ValueError:
        raise InvalidCommandGuild(
            f"Command with call `{func.__name__}` its `guilds` parameter "
            "contains a non valid guild id."
        )

    if description and len(description) > 100:
        raise CommandDescriptionTooLong(
            f"Command `{cmd}` (`{func.__name__}`) its description exceeds "
            "the 100 character limit."
        )

    group = MISSING
    sub_group = MISSING

    if isinstance(parent, Group):
        group = parent
    if isinstance(parent, Subgroup):
        group = parent.parent
        sub_group = parent

    if reg := ChatCommandHandler.register.get(
        _hash_app_command_params(cmd, guild, app_command_type, group, sub_group)
    ):
        raise CommandAlreadyRegistered(
            f"Command `{cmd}` (`{func.__name__}`) has already been "
            f"registered by `{reg.call.__name__}`."
        )

    _log.info(f"Registered command `{cmd}` to `{func.__name__}` locally.")

    interactable = InteractableStructure(
        call=func,
        cooldown=cooldown,
        cooldown_scale=cooldown_scale,
        cooldown_scope=cooldown_scope,
        manager=None,
        group=group,
        sub_group=sub_group,
        metadata=AppCommand(
            name=cmd,
            description=description,
            type=app_command_type,
            default_permission=enable_default,
            options=command_options,
            guild_id=guild_id,
        ),
    )

    ChatCommandHandler.register[
        _hash_app_command_params(
            cmd, guild_id, app_command_type, group, sub_group
        )
    ] = interactable

    return interactable



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .action_row import ActionRow
from .button import Button, ButtonStyle, LinkButton
from .component_handler import ComponentHandler
from .decorators import component, button, select_menu
from .select_menu import SelectMenu, SelectOption

__all__ = (
    "ActionRow",
    "Button",
    "ButtonStyle",
    "ComponentHandler",
    "LinkButton",
    "SelectMenu",
    "SelectOption",
    "button",
    "component",
    "select_menu",
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a user adds a reaction to a message"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects import GuildMember, Emoji
from ..objects.events.message import MessageReactionAddEvent

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def message_reaction_add_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_message_reaction_add`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the message reaction add event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.events.message.MessageReactionAddEvent`]
        ``on_message_reaction_add`` and an ``MessageReactionAddEvent``
    """  # noqa: E501

    return (
        "on_message_reaction_add",
        MessageReactionAddEvent.from_dict(
            {
                "member": GuildMember.from_dict(payload.data.pop("member")),
                "emoji": Emoji.from_dict(payload.data.pop("emoji")),
                **payload.data,
            }
        ),
    )


def export():
    return message_reaction_add_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import copy
import logging
from dataclasses import fields, _is_dataclass_instance
from enum import Enum, EnumMeta
from inspect import getfullargspec
from itertools import chain
from typing import (
    Dict,
    Tuple,
    Union,
    Generic,
    TypeVar,
    Any,
    TYPE_CHECKING,
    get_type_hints,
    get_origin,
    get_args,
    Optional,
)

from .types import MissingType, MISSING, TypeCache
from ..exceptions import InvalidArgumentAnnotation

if TYPE_CHECKING:
    from ..core import HTTPClient
    from ..objects.guild.channel import Channel
    from ..objects.guild.guild import Guild
    from ..client import Client

T = TypeVar("T")

_log = logging.getLogger(__package__)


def _asdict_ignore_none(obj: Generic[T]) -> Union[Tuple, Dict, T]:
    """
    Returns a dict from a dataclass that ignores
    all values that are None
    Modification of _asdict_inner from dataclasses

    Parameters
    ----------

    obj: Generic[T]
        The object to convert

    Returns
    -------
        A dict without None values
    """

    if _is_dataclass_instance(obj):
        result = []
        for f in fields(obj):
            value = _asdict_ignore_none(getattr(obj, f.name))

            if isinstance(value, Enum):
                result.append((f.name, value.value))
            # This if statement was added to the function
            elif not isinstance(value, MissingType) and not f.name.startswith(
                "_"
            ):
                result.append((f.name, value))

        return dict(result)

    elif isinstance(obj, tuple) and hasattr(obj, "_fields"):
        return type(obj)(*[_asdict_ignore_none(v) for v in obj])

    elif isinstance(obj, (list, tuple)):
        return type(obj)(_asdict_ignore_none(v) for v in obj)

    elif isinstance(obj, dict):
        return type(obj)(
            (_asdict_ignore_none(k), _asdict_ignore_none(v))
            for k, v in obj.items()
        )
    else:
        return copy.deepcopy(obj)


class APIObject:
    """
    Represents an object which has been fetched from the Discord API.
    """

    _client: Optional[Client] = None

    @property
    def _http(self) -> HTTPClient:
        if not self._client:
            raise AttributeError("Object is not yet linked to a client")

        return self._client.http

    @classmethod
    def bind_client(cls, client: Client):
        """
        Links the object to the client.

        Parameters
        ----------
        client: Client
            The client to link to.
        """
        cls._client = client

    def __get_types(self, attr: str, arg_type: type) -> Tuple[type]:
        """Get the types from type annotations.

        Parameters
        ----------
        attr: :class:`str`
            The attribute the typehint belongs to.
        arg_type: :class:`type`
            The type annotation for the attribute.

        Returns
        -------
        Tuple[:class:`type`]
            A collection of type annotation(s). Will most of the time
            consist of 1 item.

        Raises
        ------
        :class:`~pincer.exceptions.InvalidArgumentAnnotation`
            Exception which is raised when the type annotation has not enough
            or too many arguments for the parser to handle.
        """
        origin = get_origin(arg_type)

        if origin is Union:
            # Ahh yes, typing module has no type annotations for this...
            # noinspection PyTypeChecker
            args: Tuple[type] = get_args(arg_type)

            if 2 <= len(args) < 4:
                return args

            raise InvalidArgumentAnnotation(
                f"Attribute `{attr}` in `{type(self).__name__}` has too many "
                f"or not enough arguments! (got {len(args)} expected 2-3)"
            )

        return (arg_type,)

    def __attr_convert(self, attr_value: Dict, attr_type: T) -> T:
        """Convert an attribute to the requested attribute type using
        the factory or the __init__.

        Parameters
        ----------
        attr: :class:`str`
            The attribute the typehint belongs to.
        arg_type: T
            The type annotation for the attribute.

        Returns
        -------
        T
            The instantiated version of the arg_type.
        """
        factory = attr_type

        # Always use `__factory__` over __init__
        if getattr(attr_type, "__factory__", None):
            factory = attr_type.__factory__

        if attr_value is MISSING:
            return MISSING

        if attr_type is not None and isinstance(attr_value, attr_type):
            return attr_value

        if isinstance(attr_value, dict):
            return factory(attr_value)

        return factory(attr_value)

    def __post_init__(self):
        TypeCache()

        attributes = chain.from_iterable(
            get_type_hints(cls, globalns=TypeCache.cache).items()
            for cls in chain(self.__class__.__bases__, (self,))
        )

        for attr, attr_type in attributes:
            # Ignore private attributes.
            if attr.startswith("_"):
                continue

            types = self.__get_types(attr, attr_type)

            types = tuple(
                filter(
                    lambda tpe: tpe is not None and tpe is not MISSING, types
                )
            )

            if not types:
                raise InvalidArgumentAnnotation(
                    f"Attribute `{attr}` in `{type(self).__name__}` only "
                    "consisted of missing/optional type!"
                )

            specific_tp = types[0]

            attr_gotten = getattr(self, attr)

            if tp := get_origin(specific_tp):
                specific_tp = tp

            if isinstance(specific_tp, EnumMeta) and not attr_gotten:
                attr_value = MISSING
            elif tp == list and attr_gotten and (classes := get_args(types[0])):
                attr_value = [
                    self.__attr_convert(attr_item, classes[0])
                    for attr_item in attr_gotten
                ]
            elif tp == dict and attr_gotten and (classes := get_args(types[0])):
                attr_value = {
                    key: self.__attr_convert(value, classes[1])
                    for key, value in attr_gotten.items()
                }
            else:
                attr_value = self.__attr_convert(attr_gotten, specific_tp)

            setattr(self, attr, attr_value)

    # Set default factory method to from_dict for APIObject's.
    @classmethod
    def __factory__(cls: Generic[T], *args, **kwargs) -> T:
        return cls.from_dict(*args, **kwargs)

    def __repr__(self):
        attrs = ", ".join(
            f"{k}={v!r}"
            for k, v in self.__dict__.items()
            if v and not k.startswith("_")
        )

        return f"{type(self).__name__}({attrs})"

    def __str__(self):
        # TODO: fix docs
        """

        Returns
        -------

        """
        if _name := getattr(self, "__name__", None):
            return f"{_name} {self.__class__.__name__.lower()}"

        return super().__str__()

    @classmethod
    def from_dict(
        cls: Generic[T], data: Dict[str, Union[str, bool, int, Any]]
    ) -> T:
        """
        Parse an API object from a dictionary.
        """
        if isinstance(data, cls):
            return data

        # Disable inspection for IDE because this is valid code for the
        # inherited classes:
        # noinspection PyArgumentList
        return cls(
            **dict(
                map(
                    lambda key: (
                        key,
                        data[key].value
                        if isinstance(data[key], Enum)
                        else data[key],
                    ),
                    filter(
                        lambda object_argument: data.get(object_argument)
                        is not None,
                        getfullargspec(cls.__init__).args,
                    ),
                )
            )
        )

    def to_dict(self) -> Dict:
        """
        Transform the current object to a dictionary representation. Parameters that
        start with an underscore are not serialized.
        """
        return _asdict_ignore_none(self)


class GuildProperty:
    @property
    def guild(self) -> Guild:
        """Return a guild from an APIObject
        Parameters
        ----------
        self : :class:`~pincer.utils.api_object.APIObject`

        Returns
        -------
        :class:`~pincer.objects.guild.guild.Guild`
        """
        return self._client.guilds[self.guild_id]


class ChannelProperty:
    @property
    def channel(self) -> Channel:
        """Return a channel from an APIObject
        Parameters
        ----------
        obj : :class:`~pincer.utils.api_object.APIObject`

        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
        """
        return self._client.channels[self.channel_id]



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.


from __future__ import annotations

from asyncio import create_task, Task, ensure_future, sleep
from dataclasses import dataclass
from datetime import datetime
from itertools import repeat, count, chain
import logging
from platform import system
from random import random
from typing import TYPE_CHECKING, Any, Dict, Callable, Optional
from zlib import decompressobj

from aiohttp import (
    ClientSession,
    WSMsgType,
    ClientConnectorError,
    ClientWebSocketResponse,
)

from . import __package__
from ..utils.api_object import APIObject
from .._config import GatewayConfig
from ..core.dispatch import GatewayDispatch
from ..exceptions import (
    InvalidTokenError,
    GatewayConnectionError,
    GatewayError,
    UnhandledException,
)

if TYPE_CHECKING:
    from ..objects.app.intents import Intents

    Handler = Callable[[GatewayDispatch], None]

_log = logging.getLogger(__package__)

ZLIB_SUFFIX = b"\x00\x00\xff\xff"
inflator = decompressobj()


@dataclass
class SessionStartLimit(APIObject):
    """Session start limit info returned from the `gateway/bot` endpoint"""

    total: int
    remaining: int
    reset_after: int
    max_concurrency: int


@dataclass
class GatewayInfo(APIObject):
    """Gateway info returned from the `gateway/bot` endpoint"""

    url: str
    shards: int
    session_start_limit: SessionStartLimit


class Gateway:
    """The Gateway handles all interactions with the Discord Websocket API.
    This also contains the main event loop, and handles the heartbeat.

    Running the Gateway will create a connection with the
    Discord Websocket API on behalf of the provided token.

    This token must be a bot token.
    (Which can be found on
    `<https://discord.com/developers/applications/>`_)

    Parameters
    ----------
    token : str.
        The token for this bot
    intents : :class:`~pincer.objects.app.intents.Intents`
        The intents to use. More information can be found at
        `<https://discord.com/developers/docs/topics/gateway#gateway-intents>`_.
    url : str
        The gateway url.
    shard : int
        The ID of the shard to run.
    num_shards : int
        Number used to route traffic to the current. This should usually be the total
        number of shards that will be run. More information at
        `<https://discord.com/developers/docs/topics/gateway#sharding>`_.
    """

    def __init__(
        self,
        token: str,
        *,
        intents: Intents,
        url: str,
        shard: int,
        num_shards: int,
    ) -> None:
        self.token = token
        self.intents = intents
        self.url = url
        self.shard = shard
        self.num_shards = num_shards
        self.shard_key = [shard, num_shards]

        self.__dispatch_handlers: Dict[int, Handler] = {
            1: self.handle_heartbeat_req,
            7: self.handle_reconnect,
            9: self.handle_invalid_session,
            10: self.identify_and_handle_hello,
            11: self.handle_heartbeat,
        }

        # 4000 and 4009 are not included. The client will reconnect when receiving
        # either. Code 4000 is also used for internal disconnects that will lead to a
        # reconnect.
        self.__close_codes: Dict[int, GatewayError] = {
            4001: GatewayError("Invalid opcode was sent"),
            4002: GatewayError("Invalid payload was sent."),
            4003: GatewayError("Payload was sent prior to identifying"),
            4004: GatewayError("Token is not valid"),
            4005: GatewayError(
                "Authentication was sent after client already authenticated"
            ),
            4007: GatewayError(
                "Invalid sequence sent when starting new session"
            ),
            4008: GatewayError("Client was rate limited"),
            4010: GatewayError("Invalid shard"),
            4011: GatewayError("Sharding required"),
            4012: GatewayError("Invalid API version"),
            4013: GatewayError("Invalid intents"),
            4014: GatewayError("Disallowed intents"),
        }

        # ClientSession to be used for this Dispatcher
        self.__session: Optional[ClientSession] = None

        # This type `_WSRequestContextManager` isn't exposed by aiohttp.
        # `ClientWebSocketResponse` is a parent class.
        self.__socket: Optional[ClientWebSocketResponse] = None

        # Buffer used to store information in transport compression.
        self.__buffer = bytearray()

        # The gateway can be disconnected from Discord. This variable stores if the
        # gateway should send a hello or reconnect.
        self.__should_resume: bool = False

        # The sequence number for the last received payload. This is used reconnecting.
        self.__sequence_number: int = 0

        # The heartbeat task
        self.__heartbeat_task: Optional[Task] = None

        # Keeps the Client waiting until the next heartbeat
        self.__wait_for_heartbeat: Optional[Task] = None

        # How long the client should wait between each Heartbeat.
        self.__heartbeat_interval: Optional[int] = None

        # Tracks whether the gateway has received an ack (opcode 11) since the last
        # heartbeat.
        #   True: An ack has been received
        #   False: No ack has been received. Attempt to reconnect with gateway,
        self.__has_received_ack: bool = True

        # Session ID received from `on_ready` event. It is set in the `on_ready`
        # middleware. This is used reconnecting.
        self.__session_id: Optional[str] = None

    def __del__(self):
        """Delete method ensures all connections are closed"""
        if self.__socket:
            create_task(self.__socket.close())
        if self.__session:
            create_task(self.__session.close())

    async def init_session(self):
        """|coro|
        Crates the ClientSession. ALWAYS run this function right after initializing
        a Gateway.
        """
        self.__session = ClientSession()

    def append_handlers(self, handlers: Dict[int, Handler]):
        """The Client that uses the handler can append their own methods. The gateway
        will run those methods when the specified opcode is received.
        """
        self.__dispatch_handlers = {**self.__dispatch_handlers, **handlers}

    def set_session_id(self, _id: str):
        """Session id is private for consistency"""
        self.__session_id = _id

    def decompress_msg(self, msg: bytes) -> Optional[str]:
        if GatewayConfig.compression == "zlib-payload":
            return inflator.decompress(msg)

        if GatewayConfig.compression == "zlib-stream":
            self.__buffer.extend(msg)

            if len(self.__buffer) < 4 or self.__buffer[-4:] != ZLIB_SUFFIX:
                return None

            msg = inflator.decompress(msg)
            self.__buffer = bytearray()
            return msg

        return None

    async def start_loop(self):
        """|coro|
        Instantiate the dispatcher, this will create a connection to the
        Discord websocket API on behalf of the client whose token has
        been passed.
        """
        for _try in count():
            try:
                self.__socket = await self.__session.ws_connect(
                    GatewayConfig.make_uri(self.url)
                )
                break
            except ClientConnectorError as e:
                if _try > GatewayConfig.MAX_RETRIES:
                    raise GatewayConnectionError from e

                _log.warning(
                    "%s Could not open websocket with Discord."
                    " Retrying in 15 seconds...",
                    self.shard_key,
                )
                await sleep(15)

        _log.debug("%s Starting event loop...", self.shard_key)
        await self.event_loop()

    async def event_loop(self):
        """|coro|
        Handles receiving messages and decompressing them if needed
        """
        async for msg in self.__socket:
            if msg.type == WSMsgType.TEXT:
                await self.handle_data(msg.data)
            elif msg.type == WSMsgType.BINARY:
                # Message from transport compression that isn't complete returns None
                data = self.decompress_msg(msg.data)
                if data:
                    await self.handle_data(data)
            elif msg.type == WSMsgType.ERROR:
                raise GatewayError from self.__socket.exception()

        # The loop is broken when the gateway stops receiving messages.
        # The "error" op codes are in `self.__close_codes`. The rest of the
        # close codes are unknown issues (such as an unintended disconnect) so the
        # client should reconnect to the gateway.
        err = self.__close_codes.get(self.__socket.close_code)

        if err:
            raise err

        _log.debug(
            "%s Disconnected from Gateway due without any errors. Reconnecting.",
            self.shard_key,
        )
        # ensure_future prevents a stack overflow
        ensure_future(self.start_loop())

    async def handle_data(self, data: Dict[Any]):
        """|coro|
        Method is run when a payload is received from the gateway.
        The message is expected to already have been decompressed.
        Handling the opcode is forked to the background, so they aren't blocking.
        """
        payload = GatewayDispatch.from_string(data)

        # Op code -1 is activated on all payloads
        op_negative_one = self.__dispatch_handlers.get(-1)
        if op_negative_one:
            ensure_future(op_negative_one(payload))

        _log.debug(
            "%s %s GatewayDispatch with opcode %s received",
            self.shard_key,
            datetime.now(),
            payload.op,
        )

        # Many events are sent with a `null` sequence. This sequence should not
        # be tracked.
        if payload.seq is not None:
            self.__sequence_number = payload.seq
            _log.debug(
                "%s Set sequence number to %s", self.shard_key, payload.seq
            )

        handler = self.__dispatch_handlers.get(payload.op)

        if handler is None:
            raise UnhandledException(
                f"Opcode {payload.op} does not have a handler"
            )

        ensure_future(handler(payload))

    async def handle_heartbeat_req(self, payload: GatewayDispatch):
        """|coro|
        Opcode 1 - Instantly send a heartbeat.
        """
        self.send_next_heartbeat()

    async def handle_reconnect(self, payload: GatewayDispatch):
        """|coro|
        Opcode 7 - Reconnect and resume immediately.
        """
        _log.debug(
            "%s Requested to reconnect to Discord. Closing session and attempting to"
            " resume...",
            self.shard_key,
        )

        self.__should_resume = True
        await self.__socket.close()

    async def handle_invalid_session(self, payload: GatewayDispatch):
        """|coro|
        Opcode 9 - Invalid connection
        Attempt to relog. This is probably because the session was already invalidated
        when we tried to reconnect.
        """

        _log.debug(
            "%s Invalid session, attempting to %s...",
            self.shard_key,
            "reconnect" if payload.data else "relog",
        )

        self.__should_resume = payload.data
        self.stop_heartbeat()
        await self.__socket.close()

    async def identify_and_handle_hello(self, payload: GatewayDispatch):
        """|coro|
        Opcode 10 - Hello there general kenobi
        Runs when we connect to the gateway for the first time and every time after.
        If the client thinks it should reconnect, the opcode 6 resume payload is sent
        instead of the opcode 2 hello payload. A new session is only started after a
        reconnect if pcode 9 is received.

        Successful reconnects are handled in the `resumed` middleware.
        """
        if self.__should_resume:
            _log.debug("%s Resuming connection with Discord", self.shard_key)

            await self.send(
                str(
                    GatewayDispatch(
                        6,
                        {
                            "token": self.token,
                            "session_id": self.__session_id,
                            "seq": self.__sequence_number,
                        },
                    )
                )
            )
            return

        await self.send(
            str(
                GatewayDispatch(
                    2,
                    {
                        "token": self.token,
                        "intents": self.intents,
                        "properties": {
                            "$os": system(),
                            "$browser": __package__,
                            "$device": __package__,
                        },
                        "compress": GatewayConfig.compressed(),
                        "shard": self.shard_key,
                    },
                )
            )
        )
        self.__heartbeat_interval = payload.data["heartbeat_interval"]

        self.start_heartbeat()

    async def handle_heartbeat(self, payload: GatewayDispatch):
        """|coro|
        Opcode 11 - Heartbeat
        Track that the heartbeat has been received using shared state (Rustaceans would
        be very mad)
        """
        self.__has_received_ack = True

    async def send(self, payload: str):
        """|coro|
        Send a string object to the payload. Most of this method is just logging,
        the last line is the only one that matters for functionality.
        """
        safe_payload = payload.replace(self.token, "%s..." % self.token[:10])

        if self.__session_id:
            safe_payload = safe_payload.replace(
                self.__session_id, "%s..." % self.__session_id[:4]
            )

        _log.debug("%s Sending payload: %s", self.shard_key, safe_payload)

        if self.__socket.closed:
            _log.debug(
                "%s Socket is closing. Payload not sent.", self.shard_key
            )
            return

        await self.__socket.send_str(payload)

    def start_heartbeat(self):
        """
        Starts the heartbeat if it is not already running.
        """
        if not self.__heartbeat_task or self.__heartbeat_task.cancelled():
            self.__heartbeat_task = ensure_future(self.__heartbeat_loop())

    def stop_heartbeat(self):
        self.__heartbeat_task.cancel()

    def send_next_heartbeat(self):
        """
        It is expected to always be waiting for a heartbeat. By canceling that task,
        a heartbeat can be sent.
        """
        self.__wait_for_heartbeat.cancel()

    async def __heartbeat_loop(self):
        """|coro|
        The heartbeat is responsible for keeping the connection to Discord alive.

        Jitter is only random for the first heartbeat. It should be 1 every other
        heartbeat.
        """
        _log.debug("%s Starting heartbeat loop...", self.shard_key)

        # When waiting for first heartbeat, there hasn't been an ack received yet.
        # Set to true so the ack received check doesn't incorrectly fail.
        self.__has_received_ack = True

        for jitter in chain((random(),), repeat(1)):
            duration = self.__heartbeat_interval * jitter

            _log.debug(
                "%s %s sending heartbeat in %sms",
                self.shard_key,
                datetime.now(),
                duration,
            )

            # Task is needed so waiting can be cancelled by op code 1
            self.__wait_for_heartbeat = create_task(sleep(duration / 1000))

            await self.__wait_for_heartbeat

            if not self.__has_received_ack:
                # Close code is specified to be anything that is not 1000 in the docs.
                _log.debug(
                    "%s %s ack not received. Attempting to reconnect."
                    " Closing socket with close code 4000. \U0001f480",
                    datetime.now(),
                    self.shard_key,
                )
                await self.__socket.close(code=4000)
                self.__should_resume = True
                # A new loop is started in the background while this one is stopped.
                self.stop_heartbeat()
                return

            self.__has_received_ack = False
            await self.send(
                str(GatewayDispatch(1, data=self.__sequence_number))
            )
            _log.debug("%s sent heartbeat", self.shard_key)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a user in a subscribed voice channel speaks"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.voice import SpeakingStartEvent
from ..utils.types import Coro

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def speaking_start_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_speaking_start`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the speaking start event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`SpeakingStartEvent`]
        ``on_speaking_start`` and a ``SpeakingStartEvent``
    """
    return ("on_speaking_start", SpeakingStartEvent.from_dict(payload.data))


def export() -> Coro:
    return speaking_start_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
Sent in response to Guild Request Members. You can use the ``chunk_index``
and ``chunk_count`` to calculate how many chunks are left for your request.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.guild import GuildMembersChunkEvent
from ..utils import Coro

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def guild_member_chunk_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_guild_member_chunk`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the guild member chunk event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.events.guild.GuildMembersChunkEvent`]
        ``on_guild_member_chunk`` and a ``GuildMembersChunkEvent``
    """  # noqa: E501

    return (
        "on_guild_member_chunk",
        GuildMembersChunkEvent.from_dict(payload.data),
    )


def export() -> Coro:
    return guild_member_chunk_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .audit_log import (
    AuditLogEvent,
    AuditLogChange,
    AuditEntryInfo,
    AuditLogEntry,
    AuditLog,
)
from .ban import Ban
from .channel import (
    ChannelType,
    Channel,
    TextChannel,
    VoiceChannel,
    CategoryChannel,
    NewsChannel,
    ChannelMention,
    PrivateThread,
    PublicThread,
)
from .features import GuildFeature
from .followed_channel import FollowedChannel
from .guild import (
    PremiumTier,
    GuildNSFWLevel,
    ExplicitContentFilterLevel,
    MFALevel,
    VerificationLevel,
    DefaultMessageNotificationLevel,
    SystemChannelFlags,
    Guild,
    UnavailableGuild,
)
from .invite import InviteTargetType, InviteStageInstance, Invite
from .member import GuildMember, PartialGuildMember, BaseMember
from .overwrite import Overwrite
from .permissions import Permissions
from .role import RoleTags, Role
from .scheduled_events import (
    GuildScheduledEventEntityType,
    GuildScheduledEventUser,
    EventStatus,
    ScheduledEvent,
)
from .stage import PrivacyLevel, StageInstance
from .template import GuildTemplate
from .thread import ThreadMetadata, ThreadMember
from .webhook import WebhookType, Webhook
from .welcome_screen import WelcomeScreenChannel, WelcomeScreen
from .widget import GuildWidget


__all__ = (
    "",
    "AuditEntryInfo",
    "AuditLog",
    "AuditLogChange",
    "AuditLogEntry",
    "AuditLogEvent",
    "Ban",
    "BaseMember",
    "CategoryChannel",
    "Channel",
    "ChannelMention",
    "ChannelType",
    "DefaultMessageNotificationLevel",
    "EventStatus",
    "ExplicitContentFilterLevel",
    "FollowedChannel",
    "Guild",
    "GuildFeature",
    "GuildMember",
    "GuildNSFWLevel",
    "GuildScheduledEventEntityType",
    "GuildTemplate",
    "GuildWidget",
    "Invite",
    "InviteStageInstance",
    "InviteTargetType",
    "MFALevel",
    "NewsChannel",
    "Overwrite",
    "PartialGuildMember",
    "Permissions",
    "PremiumTier",
    "PrivacyLevel",
    "Role",
    "RoleTags",
    "ScheduledEvent",
    "StageInstance",
    "SystemChannelFlags",
    "TextChannel",
    "ThreadMember",
    "ThreadMetadata",
    "UnavailableGuild",
    "VerificationLevel",
    "VoiceChannel",
    "Webhook",
    "WebhookType",
    "WelcomeScreen",
    "WelcomeScreenChannel",
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
non-subscription event sent immediately after connecting,
contains server information
"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Tuple
    from ..utils.types import Coro
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch

_log = logging.getLogger(__package__)


async def on_resumed(
    self: Client, gateway: Gateway, payload: GatewayDispatch
) -> Tuple[str]:
    """|coro|

    Middleware for the ``on_resumed`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the stage instance create event
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`]
        ``on_ready``
    """

    _log.debug(
        "%s Successfully reconnected to Discord gateway. Restarting heartbeat.",
        gateway.shard_key,
    )
    gateway.start_heartbeat()

    return ("on_resumed",)


def export() -> Coro:
    return on_resumed



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a guild is created/joined on the client"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.guild import Guild

if TYPE_CHECKING:
    from typing import Tuple
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def guild_create_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_guild_create``,
        generate the guild class that was created

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the guild create event
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.guild.guild.Guild`]

        ``on_guild_create`` and a ``Guild``
    """
    guild = Guild.from_dict(payload.data)
    self.guilds[guild.id] = guild
    for channel in guild.channels:
        self.channels[channel.id] = channel

    return "on_guild_create", guild


def export():
    return guild_create_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from asyncio import sleep, ensure_future
from dataclasses import dataclass
from enum import IntEnum
from typing import AsyncIterator, overload, TYPE_CHECKING

from .invite import Invite, InviteTargetType
from ..message.user_message import UserMessage
from ..._config import GatewayConfig
from ...utils.api_data import APIDataGen
from ...utils.api_object import APIObject, GuildProperty
from ...utils.convert_message import convert_message
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import Dict, List, Optional, Union

    from .member import GuildMember
    from .overwrite import Overwrite
    from .thread import ThreadMetadata
    from .webhook import Webhook
    from ..message.message import Message
    from ..message.embed import Embed
    from ..user.user import User
    from ...client import Client
    from ...objects import ThreadMember
    from ...utils.timestamp import Timestamp
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake


class ChannelType(IntEnum):
    """Represents a channel its type.

    Attributes
    ----------
    GUILD_TEXT:
        A text channel.
    DM:
        A DM channel.
    GUILD_VOICE:
        A voice channel.
    GROUP_DM:
        A group DM channel.
    GUILD_CATEGORY:
        A category channel.
    GUILD_NEWS:
        A news channel.
    GUILD_STORE:
        A store channel.
    GUILD_NEWS_THREAD:
        A news thread.
    GUILD_PUBLIC_THREAD:
        A public thread.
    GUILD_PRIVATE_THREAD:
        A private thread.
    GUILD_STAGE_VOICE:
        A stage channel.
    """

    GUILD_TEXT = 0
    DM = 1
    GUILD_VOICE = 2
    GROUP_DM = 3
    GUILD_CATEGORY = 4
    GUILD_NEWS = 5
    GUILD_STORE = 6

    if GatewayConfig.version >= 9:
        GUILD_NEWS_THREAD = 10
        GUILD_PUBLIC_THREAD = 11
        GUILD_PRIVATE_THREAD = 12

    GUILD_STAGE_VOICE = 13


@dataclass(repr=False)
class Channel(APIObject, GuildProperty):  # noqa E501
    """Represents a Discord Channel Mention object

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of this channel
    type: :class:`~pincer.objects.guild.channel.ChannelType`
        The type of channel
    application_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Application id of the group DM creator if it is bot-created
    bitrate: APINullable[:class:`int`]
        The bitrate (in bits) of the voice channel
    default_auto_archive_duration: APINullable[:class:`int`]
        Default duration for newly created threads, in minutes, to
        automatically archive the thread after recent activity, can be set to:
        60, 1440, 4320, 10080
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the guild (maybe missing for some channel objects received
        over gateway guild dispatches)
    icon: APINullable[Optional[:class:`str`]]
        Icon hash
    last_message_id: APINullable[Optional[:class:`~pincer.utils.snowflake.Snowflake`]]
        The id of the last message sent in this channel (may not point to an
        existing or valid message)
    last_pin_timestamp: APINullable[Optional[:class:`~pincer.utils.timestamp.Timestamp`]]
        When the last pinned message was pinned. This may be null in events
        such as GUILD_CREATE when a message is not pinned.
    member: APINullable[:class:`~pincer.objects.guild.member.GuildMember`]
        Thread member object for the current user, if they have joined the
        thread, only included on certain API endpoints
    member_count: APINullable[:class:`int`]
        An approximate count of users in a thread, stops counting at 50
    message_count: :class:`int`
        An approximate count of messages in a thread, stops counting at 50
    name: APINullable[:class:`str`]
        The name of the channel (1-100 characters)
    nsfw: APINullable[:class:`bool`]
        Whether the channel is nsfw
    owner_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of the creator of the group DM or thread
    parent_id: APINullable[Optional[:class:`~pincer.utils.snowflake.Snowflake`]]
        For guild channels: id of the parent category for a channel (each
        parent category can contain up to 50 channels), for threads: id of the
        text channel this thread was created
    permissions: APINullable[:class:`str`]
        Computed permissions for the invoking user in the channel, including
        overwrites, only included when part of the resolved data received on a
        slash command interaction
    permission_overwrites: APINullable[List[:class:`~pincer.objects.guild.overwrite.Overwrite`]]
        Explicit permission overwrites for members and roles
    position: APINullable[:class:`int`]
        Sorting position of the channel
    rate_limit_per_user: APINullable[:class:`int`]
        Amount of seconds a user has to wait before sending another message
        (0-21600); bots, as well as users with the permission manage_messages
        or manage_channel, are unaffected
    recipients: APINullable[List[:class:`~pincer.objects.user.user.User`]]
        The recipients of the DM
    rtc_region: APINullable[Optional[:class:`str`]]
        Voice region id for the voice channel, automatic when set to null
    thread_metadata: APINullable[:class:`~pincer.objects.guild.thread.ThreadMetadata`]
        Thread-specific fields not needed by other channels
    topic: APINullable[Optional[:class:`str`]]
        The channel topic (0-1024 characters)
    user_limit: APINullable[:class:`int`]
        The user limit of the voice channel
    video_quality_mode: APINullable[:class:`int`]
        The camera video quality mode of the voice channel, 1 when not present
    """

    # noqa: E501
    id: Snowflake
    type: ChannelType

    application_id: APINullable[Snowflake] = MISSING
    bitrate: APINullable[int] = MISSING
    default_auto_archive_duration: APINullable[int] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    icon: APINullable[Optional[str]] = MISSING

    last_message_id: APINullable[Optional[Snowflake]] = MISSING
    last_pin_timestamp: APINullable[Optional[Timestamp]] = MISSING
    member: APINullable[GuildMember] = MISSING

    member_count: APINullable[int] = MISSING

    message_count: APINullable[int] = MISSING

    name: APINullable[str] = MISSING
    nsfw: APINullable[bool] = MISSING
    owner_id: APINullable[Snowflake] = MISSING
    parent_id: APINullable[Optional[Snowflake]] = MISSING
    permissions: APINullable[str] = MISSING
    permission_overwrites: APINullable[List[Overwrite]] = MISSING
    # Position is always 0 when not sent
    position: APINullable[int] = 0
    rate_limit_per_user: APINullable[int] = MISSING
    recipients: APINullable[List[User]] = MISSING
    rtc_region: APINullable[Optional[str]] = MISSING
    thread_metadata: APINullable[ThreadMetadata] = MISSING
    topic: APINullable[Optional[str]] = MISSING
    user_limit: APINullable[int] = MISSING
    video_quality_mode: APINullable[int] = MISSING

    @property
    def mention(self):
        return f"<#{self.id}>"

    @classmethod
    async def from_id(cls, client: Client, channel_id: int) -> Channel:
        """|coro|
        Creates a channel object. You should use the ``get_channel`` method
        from :class:`~pincer.client.Client` most of the time. The
        ``get_dm_channel`` method from :class:`~pincer.objects.user.user.User`
        should be used if you need to create a dm_channel. Using the ``send()``
        method from :class:`~pincer.objects.user.user.User` is preferred.

        Parameters
        ----------
        client : :class:`~pincer.client.Client`
            Client object to use the HTTP class of.
        channel_id : :class:`int`
            ID of the channel you want.

        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
            The channel object.
        """
        data = (await client.http.get(f"channels/{channel_id}")) or {}

        data.update({"type": ChannelType(data.pop("type"))})

        channel_cls = _channel_type_map.get(data["type"], Channel)
        return channel_cls.from_dict(data)

    @overload
    async def edit(
        self,
        *,
        name: str = None,
        type: ChannelType = None,
        position: int = None,
        topic: str = None,
        nsfw: bool = None,
        rate_limit_per_user: int = None,
        bitrate: int = None,
        user_limit: int = None,
        permissions_overwrites: List[Overwrite] = None,
        parent_id: Snowflake = None,
        rtc_region: str = None,
        video_quality_mod: int = None,
        default_auto_archive_duration: int = None,
    ) -> Channel:
        ...

    async def edit(self, reason: Optional[str] = None, **kwargs):
        """|coro|
        Edit a channel with the given keyword arguments.

        Parameters
        ----------
        reason Optional[:class:`str`]
            The reason of the channel edit.
        \\*\\*kwargs :
            The keyword arguments to edit the channel with.

        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
            The updated channel object.
        """
        headers = {}

        if reason is not None:
            headers["X-Audit-Log-Reason"] = str(reason)

        data = await self._http.patch(
            f"channels/{self.id}", kwargs, headers=headers
        )
        data.update({"type": ChannelType(data.pop("type"))})
        channel_cls = _channel_type_map.get(data["type"], Channel)
        return channel_cls.from_dict(data)

    async def edit_permissions(
        self,
        overwrite: Overwrite,
        allow: str,
        deny: str,
        type: int,
        reason: Optional[str] = None,
    ):
        """|coro|
        Edit the channel permission overwrites for a user or role in a channel.
        Only usable for guild channels. Requires the ``MANAGE_ROLES`` permission.
        Only permissions your bot has in the guild or channel can be
        allowed/denied (unless your bot has a ``MANAGE_ROLES`` overwrite in the channel).

        Parameters
        ----------
        overwrite: :class:`~pincer.objects.guild.overwrite.Overwrite`
            The overwrite object.
        allow: :class:`str`
            The bitwise value of all allowed permissions.
        deny: :class:`str`
            The bitwise value of all denied permissions.
        type: :class:`int`
            0 for a role or 1 for a member.
        reason: Optional[:class:`str`]
            The reason of the channel permission edit.
        """
        await self._http.put(
            f"channels/{self.id}/permissions/{overwrite.id}",
            headers={"X-Audit-Log-Reason": reason},
            data={"allow": allow, "deny": deny, "type": type},
        )

    async def delete_permission(
        self, overwrite: Overwrite, reason: Optional[str] = None
    ):
        """|coro|
        Delete a channel permission overwrite for a user or role in a channel.
        Only usable for guild channels. Requires the ``MANAGE_ROLES`` permission.

        Parameters
        ----------
        overwrite: :class:`~pincer.objects.guild.overwrite.Overwrite`
            The overwrite object.
        reason: Optional[:class:`str`]
            The reason of the channel permission delete.
        """
        await self._http.delete(
            f"channels/{self.id}/permissions/{overwrite.id}",
            headers={"X-Audit-Log-Reason": reason},
        )

    async def follow_news_channel(
        self, webhook_channel_id: Snowflake
    ) -> NewsChannel:
        """|coro|
        Follow a News Channel to send messages to a target channel.
        Requires the ``MANAGE_WEBHOOKS`` permission in the target channel.
        Returns a followed channel object.

        Parameters
        ----------
        webhook_channel_id: :class:`Snowflake`
            The ID of the channel to follow.

        Returns
        -------
        :class:`~pincer.objects.guild.channel.NewsChannel`
            The followed channel object.
        """
        return NewsChannel.from_dict(
            self._http.post(
                f"channels/{self.id}/followers",
                data={"webhook_channel_id": webhook_channel_id},
            )
        )

    async def trigger_typing_indicator(self):
        """|coro|
        Post a typing indicator for the specified channel.
        Generally bots should **not** implement this route. However, if a bot is
        responding to a command and expects the computation to take a few
        seconds, this endpoint may be called to let the user know that the bot
        is processing their message.
        """
        await self._http.post(f"channels/{self.id}/typing")

    def get_pinned_messages(self) -> APIDataGen[UserMessage]:
        """|coro|
        Fetches all pinned messages in the channel. Returns an iterator of
        pinned messages.

        Returns
        -------
        :class:`AsyncIterator[:class:`~pincer.objects.guild.message.UserMessage`]`
            An iterator of pinned messages.
        """
        return APIDataGen(
            UserMessage, self._http.get(f"channels/{self.id}/pins")
        )

    async def pin_message(
        self, message: UserMessage, reason: Optional[str] = None
    ):
        """|coro|
        Pin a message in a channel. Requires the ``MANAGE_MESSAGES`` permission.
        The maximum number of pinned messages is ``50``.

        Parameters
        ----------
        reason: Optional[:class:`str`]
            The reason of the channel message pin.
        """
        await self._http.put(
            f"channels/{self.id}/pins/{message.id}",
            headers={"X-Audit-Log-Reason": reason},
        )

    async def unpin_message(
        self, message: UserMessage, reason: Optional[str] = None
    ):
        """|coro|
        Unpin a message in a channel. Requires the ``MANAGE_MESSAGES`` permission.

        Parameters
        ----------
        reason: Optional[:class:`str`]
            The reason of the channel message unpin.
        """
        await self._http.delete(
            f"channels/{self.id}/pins/{message.id}",
            headers={"X-Audit-Log-Reason": reason},
        )

    async def group_dm_add_recipient(
        self,
        user: User,
        access_token: Optional[str] = None,
        nick: Optional[str] = None,
    ):
        """|coro|
        Adds a recipient to a Group DM using their access token.

        Parameters
        ----------
        user: :class:`~pincer.objects.user.User`
            The user to add.
        access_token: Optional[:class:`str`]
            The access token of the user that has granted your app the
            ``gdm.join`` scope.
        nick: Optional[:class:`str`]
            The nickname of the user being added.
        """
        await self._http.put(
            f"channels/{self.id}/recipients/{user.id}",
            data={"access_token": access_token, "nick": nick},
        )

    async def group_dm_remove_recipient(self, user: User):
        """|coro|
        Removes a recipient from a Group DM.

        Parameters
        ----------
        user: :class:`~pincer.objects.user.User`
            The user to remove.
        """
        await self._http.delete(f"channels/{self.id}/recipients/{user.id}")

    async def bulk_delete_messages(
        self, messages: List[Snowflake], reason: Optional[str] = None
    ):
        """|coro|
        Delete multiple messages in a single request.
        This endpoint can only be used on guild channels and requires
        the ``MANAGE_MESSAGES`` permission.

        This endpoint will not delete messages older than 2 weeks, and will
        fail with a 400 BAD REQUEST if any message provided is older than that
        or if any duplicate message IDs are provided.

        Parameters
        ----------
        messages: List[:class:`~.pincer.utils.Snowflake`]
            The list of message IDs to delete (2-100).
        reason: Optional[:class:`str`]
            The reason of the channel bulk delete.
        """
        await self._http.post(
            f"channels/{self.id}/messages/bulk_delete",
            headers={"X-Audit-Log-Reason": reason},
            data={"messages": messages},
        )

    async def delete(
        self,
        reason: Optional[str] = None,
        /,
        channel_id: Optional[Snowflake] = None,
    ):
        """|coro|

        Delete the current channel.

        Parameters
        ----------
        reason Optional[:class:`str`]
            The reason of the channel delete.
        channel_id :class:`~.pincer.utils.Snowflake`
            The id of the channel, defaults to the current object id.
        """
        channel_id = channel_id or self.id

        await self._http.delete(
            f"channels/{channel_id}",
            headers={"X-Audit-Log-Reason": reason},
        )

    @staticmethod
    async def __post_send_handler(message: UserMessage):
        """Process a message after it was sent.

        Parameters
        ----------
        message :class:`~.pincer.objects.message.message.Message`
            The message.
        """

        if getattr(message, "delete_after", None):
            await sleep(message.delete_after)
            await message.delete()

    def __post_sent(self, message: UserMessage):
        """Ensure the `__post_send_handler` method its future.

        Parameters
        ----------
        message :class:`~.pincer.objects.message.message.Message`
            The message.
        """
        ensure_future(self.__post_send_handler(message))

    async def send(self, message: Union[Embed, Message, str]) -> UserMessage:
        """|coro|

        Send a message in the channel.

        Parameters
        ----------
        message :class:`~.pincer.objects.message.message.Message`
            The message which must be sent

        Returns
        -------
        :class:`~.pincer.objects.message.user_message.UserMessage`
            The message that was sent.
        """
        content_type, data = convert_message(self._client, message).serialize()

        resp = await self._http.post(
            f"channels/{self.id}/messages", data, content_type=content_type
        )
        msg = UserMessage.from_dict(resp)
        self.__post_sent(msg)
        return msg

    def get_webhooks(self) -> APIDataGen[Webhook]:
        """|coro|
        Get all webhooks in the channel.
        Requires the ``MANAGE_WEBHOOKS`` permission.

        Yields
        -------
        AsyncGenerator[:class:`~.pincer.objects.guild.webhook.Webhook`, None]
        """
        return APIDataGen(
            Webhook, self._http.get(f"channels/{self.id}/webhooks")
        )

    def get_invites(self) -> APIDataGen[Invite]:
        """|coro|
        Fetches all the invite objects for the channel. Only usable for
        guild channels. Requires the ``MANAGE_CHANNELS`` permission.

        Returns
        -------
        AsyncIterator[:class:`~pincer.objects.guild.invite.Invite`]
            Invites iterator.
        """
        return APIDataGen(Invite, self._http.get(f"channels/{self.id}/invites"))

    async def create_invite(
        self,
        max_age: int = 86400,
        max_uses: int = 0,
        temporary: bool = False,
        unique: bool = False,
        target_type: InviteTargetType = None,
        target_user_id: Snowflake = None,
        target_application_id: Snowflake = None,
        reason: Optional[str] = None,
    ):
        """|coro|
        Create a new invite object for the channel. Only usable for guild
        channels. Requires the ``CREATE_INSTANT_INVITE`` permission.

        Parameters
        ----------
        max_age: Optional[:class:`int`]
            Duration of invite in seconds before expiry, or 0 for never. between
            0 and 604800 (7 days).
            |default| :data:`86400`
        max_uses: Optional[:class:`int`]
            Maximum number of uses. ``0`` for unlimited. Values between 0 and 100.
            |default| :data:`0`
        temporary: Optional[:class:`bool`]
            Whether the invite only grants temporary membership.
            |default| :data:`False`
        unique: Optional[:class:`bool`]
            If ``True``, don't try to reuse a similar invite (useful for
            creating many unique one time use invites).
            |default| :data:`False`
        target_type: Optional[:class:`~.pincer.objects.guild.invite.InviteTargetType`]
            The type of target for the invite.
            |default| :data:`None`
        target_user_id: Optional[:class:`~.pincer.utils.Snowflake`]
            The id of the user whose stream to display for this invite. Required
            if ``target_type`` is ``STREAM``, the user must be streaming in the
            channel.
            |default| :data:`None`
        target_application_id: Optional[:class:`~.pincer.utils.Snowflake`]
            The id of the embedded application to open for this invite. Required
            if ``target_type`` is ``EMBEDDED_APPLICATION``, the application must
            have the ``EMBEDDED`` flag.
            |default| :data:`None`
        reason: Optional[:class:`str`]
            The reason of the invite creation.
            |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.guild.invite.Invite`
            The invite object.
        """
        return Invite.from_dict(
            await self._http.post(
                f"channels/{self.id}/invites",
                headers={"X-Audit-Log-Reason": reason},
                data={
                    "max_age": max_age,
                    "max_uses": max_uses,
                    "temporary": temporary,
                    "unique": unique,
                    "target_type": target_type,
                    "target_user_id": target_user_id,
                    "target_application_id": target_application_id,
                },
            )
        )

    async def list_active_threads(self) -> ThreadsResponse:
        """|coro|
        Returns all active threads in the channel, including public and
        private threads. Threads are ordered by their id, in descending order.

        Returns
        -------
        :class:`~pincer.objects.channel.ThreadsResponse`
            The response object.
        """
        return ThreadsResponse.from_dict(
            self._http.get(f"channels/{self.id}/threads/active")
        )

    async def list_public_archived_threads(
        self, before: Optional[Timestamp] = None, limit: Optional[int] = None
    ) -> ThreadsResponse:
        """|coro|
        Returns archived threads in the channel that are public.
        When called on a ``GUILD_TEXT`` channel, returns threads of type
        ``GUILD_PUBLIC_THREAD``. When called on a ``GUILD_NEWS`` channel returns
        threads of type ``GUILD_NEWS_THREAD``. Threads are ordered by
        ``archive_timestamp``, in descending order. Requires the
        ``READ_MESSAGE_HISTORY`` permission.

        Parameters
        ----------
        before: Optional[:class:`~pincer.objects.timestamp.Timestamp`]
            Returns threads before this timestamp.
            |default| :data:`None`
        limit: Optional[:class:`int`]
            The maximum number of threads to return.
            |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.channel.ThreadsResponse`
            The response object.
        """
        return ThreadsResponse.from_dict(
            await self._http.get(
                f"channels/{self.id}/threads/archived/public",
                params={"before": before, "limit": limit},
            )
        )

    async def list_private_archived_threads(
        self, before: Optional[Timestamp] = None, limit: Optional[int] = None
    ) -> ThreadsResponse:
        """|coro|
        Returns archived threads in the channel that are of type
        ``GUILD_PRIVATE_THREAD``. Threads are ordered by ``archive_timestamp``,
        in descending order. Requires both the ``READ_MESSAGE_HISTORY`` and
        ``MANAGE_THREADS`` permissions.

        Parameters
        ----------
        before: Optional[:class:`~pincer.objects.timestamp.Timestamp`]
            Returns threads before this timestamp.
            |default| :data:`None`
        limit: Optional[:class:`int`]
            The maximum number of threads to return.
            |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.channel.ThreadsResponse`
            The response object.
        """
        return ThreadsResponse.from_dict(
            await self._http.get(
                f"channels/{self.id}/threads/archived/private",
                params={"before": before, "limit": limit},
            )
        )

    async def list_joined_private_archived_threads(
        self, before: Optional[Timestamp] = None, limit: Optional[int] = None
    ) -> ThreadsResponse:
        """|coro|
        Returns archived threads in the channel that are of type
        ``GUILD_PRIVATE_THREAD``, and the user has joined. Threads are ordered
        by their id, in descending order. Requires the ``READ_MESSAGE_HISTORY``
        permission.

        Parameters
        ----------
        before: Optional[:class:`~pincer.objects.timestamp.Timestamp`]
            Returns threads before this timestamp.
            |default| :data:`None`
        limit: Optional[:class:`int`]
            The maximum number of threads to return.
            |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.channel.ThreadsResponse`
            The response object.
        """

        return ThreadsResponse.from_dict(
            self._http.get(
                f"channels/{self.id}/users/@me/threads/archived/private",
                params={"before": before, "limit": limit},
            )
        )

    def __str__(self):
        """return the discord tag when object gets used as a string."""
        return self.name or str(self.id)


class TextChannel(Channel):
    """A subclass of ``Channel`` for text channels with all the same attributes."""

    @overload
    async def edit(
        self,
        name: str = None,
        type: ChannelType = None,
        position: int = None,
        topic: str = None,
        nsfw: bool = None,
        rate_limit_per_user: int = None,
        permissions_overwrites: List[Overwrite] = None,
        parent_id: Snowflake = None,
        default_auto_archive_duration: int = None,
    ) -> Union[TextChannel, NewsChannel]:
        ...

    async def edit(self, **kwargs):
        """|coro|
        Edit a text channel with the given keyword arguments.

        Parameters
        ----------
        \\*\\*kwargs :
            The keyword arguments to edit the channel with.

        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
            The updated channel object.
        """
        return await super().edit(**kwargs)

    async def fetch_message(self, message_id: int) -> UserMessage:
        """|coro|
        Returns a UserMessage from this channel with the given id.

        Parameters
        ----------
        message_id : :class: int
            The message ID to look for.

        Returns
        -------
        :class:`~pincer.objects.message.user_message.UserMessage`
            The requested message.
        """
        return UserMessage.from_dict(
            await self._http.get(f"channels/{self.id}/messages/{message_id}")
        )

    async def history(
        self,
        limit: int = 50,
        before: Optional[Union[int, Snowflake]] = None,
        after: Optional[Union[int, Snowflake]] = None,
        around: Optional[Union[int, Snowflake]] = None,
    ) -> AsyncIterator[UserMessage]:
        """|coro|
        Returns a list of messages in this channel.

        Parameters
        ----------
        around : Optional[Union[:class:`int`, :class:`Snowflake`]]
            The message ID to look around.
        after : Optional[Union[:class:`int`, :class:`Snowflake`]]
            The message ID to look after.
        before : Optional[Union[:class:`int`, :class:`Snowflake`]]
            The message ID to look before.
        limit : Optional[Union[:class:`int`, :class:`Snowflake`]]
            The maximum number of messages to return.

        Returns
        -------
        AsyncIterator[:class:`~pincer.objects.message.user_message.UserMessage`]
            An iterator of messages.
        """

        if limit is None:
            limit = float("inf")

        while limit > 0:
            retrieve = min(limit, 100)

            raw_messages = await self._http.get(
                f"/channels/{self.id}/messages",
                params={
                    "limit": retrieve,
                    "before": before,
                    "after": after,
                    "around": around,
                },
            )

            if not raw_messages:
                break

            for _message in raw_messages:
                yield UserMessage.from_dict(_message)

            before = raw_messages[-1]["id"]
            limit -= retrieve


class VoiceChannel(Channel):
    """A subclass of ``Channel`` for voice channels with all the same attributes."""

    @overload
    async def edit(
        self,
        name: str = None,
        position: int = None,
        bitrate: int = None,
        user_limit: int = None,
        permissions_overwrites: List[Overwrite] = None,
        rtc_region: str = None,
        video_quality_mod: int = None,
    ) -> VoiceChannel:
        ...

    async def edit(self, **kwargs):
        """|coro|
        Edit a voice channel with the given keyword arguments.

        Parameters
        ----------
        \\*\\*kwargs :
            The keyword arguments to edit the channel with.

        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
            The updated channel object.
        """
        return await super().edit(**kwargs)


class GroupDMChannel(Channel):
    """A subclass of ``Channel`` for Group DMs"""


class CategoryChannel(Channel):
    """A subclass of ``Channel`` for categories channels
    with all the same attributes.
    """


class NewsChannel(Channel):
    """A subclass of ``Channel`` for news channels with all the same attributes."""

    @overload
    async def edit(
        self,
        name: str = None,
        type: ChannelType = None,
        position: int = None,
        topic: str = None,
        nsfw: bool = None,
        permissions_overwrites: List[Overwrite] = None,
        parent_id: Snowflake = None,
        default_auto_archive_duration: int = None,
    ) -> Union[TextChannel, NewsChannel]:
        ...

    async def edit(self, **kwargs):
        """|coro|
        Edit a news channel with the given keyword arguments.

        Parameters
        ----------
        \\*\\*kwargs :
            The keyword arguments to edit the channel with.

        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
            The updated channel object.
        """
        return await super().edit(**kwargs)


class Thread(Channel):
    """A subclass of ``Channel`` for threads with all the same attributes."""

    async def start(
        self,
        name: Optional[str] = None,
        auto_archive_duration: Optional[int] = None,
        type_: Optional[ChannelType] = None,
        invitable: Optional[bool] = None,
        rate_limit_per_user: Optional[int] = None,
        reason: Optional[str] = None,
    ) -> Channel:
        """|coro|
        Creates a new thread that is not connected to an existing message.
        The created thread defaults to a ``GUILD_PRIVATE_THREAD``*.
        Returns a Channel on success.

        Parameters
        ----------
        name: Optional[:class:`str`]
            The name of the thread. 1-100 characters.
            |default| :data:`None`
        auto_archive_duration**: Optional[:class:`int`]
            The duration in minutes to automatically archive the thread after
            recent activity, can be set to: ``60``, ``1440``, ``4320``, ``10080``.
            |default| :data:`None`
        type_: Optional[:class:`~pincer.objects.channel.ChannelType`]
            The type of thread to create.
            |default| :data:`None`
        invitable: Optional[:class:`bool`]
            Whether non-moderators can add other non-moderators to a thread;
            only available when creating a private thread.
            |default| :data:`None`
        rate_limit_per_user: Optional[:class:`int`]
            Amount of seconds a user has to wait before sending another message.
            (0-21600)
            |default| :data:`None`
        reason: Optional[:class:`str`]
            The reason of the thread creation.
            |default| :data:`None`

        \\*: Creating a private thread requires the server to be boosted.
        The guild features will indicate if that is possible for the guild.

        \\*\\*: The 3 day and 7 day archive durations require the server to be boosted.
        The guild features will indicate if that is possible for the guild.

        Returns
        -------
        :class:`~pincer.objects.channel.Channel`
            The created thread.
        """
        return Channel.from_dict(
            await self._http.post(
                f"channels/{self.id}/threads",
                headers={"X-Audit-Log-Reason": reason},
                data={
                    "name": name,
                    "auto_archive_duration": auto_archive_duration,
                    "type": type_,
                    "invitable": invitable,
                    "rate_limit_per_user": rate_limit_per_user,
                },
            ),
        )

    async def start_with_message(
        self,
        message: Optional[UserMessage],
        name: Optional[str] = None,
        auto_archive_duration: Optional[int] = None,
        rate_limit_per_user: Optional[int] = None,
        reason: Optional[str] = None,
    ) -> Channel:
        """|coro|
        Creates a new thread from an existing message. Returns a Channel on
        success.

        When called on a ``GUILD_TEXT`` channel, creates a ``GUILD_PUBLIC_THREAD``.
        When called on a ``GUILD_NEWS`` channel, creates a ``GUILD_NEWS_THREAD``.

        The id of the created thread will be the same as the id of the message,
        and as such a message can only have a single thread created from it.

        Parameters
        ----------
        message: :class:`~pincer.objects.message.user_message.UserMessage`
            The message to create the thread from.
            |default| data:`None`
        name: Optional[:class:`str`]
            The name of the thread. 1-100 characters.
            |default| data:`None`
        auto_archive_duration: Optional[:class:`int`]
            The duration in minutes to automatically archive the thread after
            recent activity, can be set to: ``60``, ``1440``, ``4320``, ``10080``.
            |default| data:`None`
        rate_limit_per_user: Optional[:class:`int`]
            Amount of seconds a user has to wait before sending another message.
            (0-21600)
            |default| data:`None`
        reason: Optional[:class:`str`]
            The reason of the thread creation.
            |default| data:`None`

        The 3 day and 7 day archive durations require the server to be boosted.
        The guild features will indicate if a server is able to use those settings.

        Returns
        -------
        :class:`~pincer.objects.channel.Channel`
            The created thread.
        """
        return Channel.from_dict(
            await self._http.post(
                f"channels/{self.id}/messages/{message.id}/threads",
                headers={"X-Audit-Log-Reason": reason},
                data={
                    "name": name,
                    "auto_archive_duration": auto_archive_duration,
                    "rate_limit_per_user": rate_limit_per_user,
                },
            )
        )

    async def join(self):
        """|coro|
        Adds the current user to a thread.
        Also requires the thread to not be archived.
        """
        await self._http.put(f"channels/{self.id}/thread-members/@me")

    async def add_member(self, user: User):
        """|coro|
        Adds another member to a thread.
        Requires the ability to send messages in the thread.
        Also requires the thread to not be archived.

        Parameters
        ----------
        user: :class:`~pincer.objects.user.User`
            The user to add to the thread.
        """
        await self._http.put(f"channels/{self.id}/thread-members/{user.id}")

    async def leave(self):
        """|coro|
        Removes the current user from a thread.
        Also requires the thread to not be archived.
        """
        await self._http.delete(f"channels/{self.id}/thread-members/@me")

    async def remove_member(self, user: User):
        """|coro|
        Removes another member from a thread. Requires the ``MANAGE_THREADS``
        permission, or the creator of the thread if it is a
        ``GUILD_PRIVATE_THREAD``. Also requires the thread to not be archived.

        Parameters
        ----------
        user: :class:`~pincer.objects.user.User`
            The user to remove from the thread.
        """
        await self._http.delete(f"channels/{self.id}/thread-members/{user.id}")

    async def get_member(self, user: User) -> ThreadMember:
        """|coro|
        Returns a thread member object for the specified user if they are a
        member of the thread.

        Parameters
        ----------
        user: :class:`~pincer.objects.user.User`
            The user to get the thread member for.

        Returns
        -------
        :class:`~pincer.objects.channel.ThreadMember`
            The thread member object.
        """
        return ThreadMember.from_dict(
            await self._http.get(f"channels/{self.id}/thread-members/{user.id}")
        )

    def list_members(self) -> APIDataGen[ThreadMember]:
        """|coro|
        Fetches all the thread members for the thread. Returns an iterator of
        ThreadMember objects.

        Returns
        -------
        AsyncIterator[:class:`~pincer.objects.channel.ThreadMember`]
            An iterator of thread members.
        """
        return APIDataGen(
            ThreadMember, self._http.get(f"channels/{self.id}/thread-members")
        )


class PublicThread(Thread):
    """A subclass of ``Thread`` for public threads with all the same attributes."""


class PrivateThread(Thread):
    """A subclass of ``Thread`` for private threads with all the same attributes."""


@dataclass(repr=False)
class ThreadsResponse(APIObject):
    """A class representing a response from the API for a list of threads."""

    threads: AsyncIterator[Thread]
    members: AsyncIterator[ThreadMember]
    has_more: bool


@dataclass(repr=False)
class ChannelMention(APIObject):
    """Represents a Discord Channel Mention object

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the channel
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the guild containing the channel
    type: :class:`~pincer.objects.guild.channel.ChannelType`
        The type of channel
    name: :class:`str`
        The name of the channel
    """

    id: Snowflake
    guild_id: Snowflake
    type: ChannelType
    name: str


# noinspection PyTypeChecker
_channel_type_map: Dict[ChannelType, Channel] = {
    ChannelType.GUILD_TEXT: TextChannel,
    ChannelType.GUILD_VOICE: VoiceChannel,
    ChannelType.GROUP_DM: GroupDMChannel,
    ChannelType.GUILD_CATEGORY: CategoryChannel,
    ChannelType.GUILD_NEWS: NewsChannel,
    ChannelType.GUILD_PUBLIC_THREAD: PublicThread,
    ChannelType.GUILD_PRIVATE_THREAD: PrivateThread,
}



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from enum import IntFlag


class InteractionFlags(IntFlag):
    """

    Attributes
    ----------
    EPHEMERAL:
        Only the user receiving the message can see it.
    SUPPRESS_EMBEDS:
        No link embeds should be shown for this message.
    """

    EPHEMERAL = 1 << 6
    SUPPRESS_EMBEDS = 1 << 2



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a message in pinned/unpinned"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.channel import ChannelPinsUpdateEvent

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def channel_pins_update_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_channel_pins_update`` event.

    Parameters
    ----------
    payload : :class:`pincer.core.gateway.GatewayDispatch`
        The data received from the channel pins update event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.guild.channel.Channel`]
        ``on_channel_pins_update`` and a ``Channel``
    """

    event = ChannelPinsUpdateEvent.from_dict(payload.data)
    self.channels[
        event.channel_id
    ].last_pin_timestamp = event.last_pin_timestamp

    return ("on_channel_pins_update", event)


def export():
    return channel_pins_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .connection import Connection
from .integration import (
    IntegrationExpireBehavior,
    IntegrationApplication,
    Integration,
    IntegrationAccount,
)
from .user import User, PremiumTypes, VisibilityType
from .voice_state import VoiceState


__all__ = (
    "Connection",
    "Integration",
    "IntegrationAccount",
    "IntegrationApplication",
    "IntegrationExpireBehavior",
    "PremiumTypes",
    "User",
    "VisibilityType",
    "VoiceState",
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .api_object import APIObject, ChannelProperty, GuildProperty
from .color import Color
from .conversion import remove_none
from .directory import chdir
from .event_mgr import EventMgr
from .extraction import get_index
from .insertion import should_pass_cls, should_pass_ctx
from .replace import replace
from .shards import calculate_shard_id
from .signature import get_params, get_signature_and_params
from .snowflake import Snowflake
from .tasks import Task, TaskScheduler
from .timestamp import Timestamp
from .types import (
    APINullable,
    Coro,
    MISSING,
    MissingType,
    choice_value_types,
    CheckFunction,
)

__all__ = (
    "APINullable",
    "APIObject",
    "ChannelProperty",
    "CheckFunction",
    "Color",
    "Coro",
    "EventMgr",
    "GuildProperty",
    "MISSING",
    "MissingType",
    "Snowflake",
    "Task",
    "TaskScheduler",
    "Timestamp",
    "calculate_shard_id",
    "chdir",
    "choice_value_types",
    "get_index",
    "get_params",
    "get_signature_and_params",
    "remove_none",
    "replace",
    "should_pass_cls",
    "should_pass_ctx",
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass


@dataclass(repr=False)
class GatewayConfig:
    """This file is to make maintaining the library and its gateway
    configuration easier. Leave compression blank for no compression.
    """

    MAX_RETRIES: int = 5
    version: int = 9
    encoding: str = "json"
    compression: str = "zlib-payload"

    @classmethod
    def make_uri(cls, uri) -> str:
        """
        Returns
        -------
        :class:`str`:
            The GatewayConfig's uri.
        """
        return (
            f"{uri}" f"?v={cls.version}" f"&encoding={cls.encoding}"
        ) + f"&compress={cls.compression}" * cls.compressed()

    @classmethod
    def compressed(cls) -> bool:
        """
        Returns
        -------
        :class:`bool`:
            Whether the Gateway should compress payloads or not.
        """
        return cls.compression in ["zlib-stream", "zlib-payload"]



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .commands import command, user_command, message_command
from .chat_command_handler import ChatCommandHandler
from .arg_types import (
    CommandArg,
    Description,
    Choice,
    Choices,
    ChannelTypes,
    MaxValue,
    MinValue,
    Modifier,
)
from .components import (
    ActionRow,
    Button,
    ButtonStyle,
    ComponentHandler,
    SelectMenu,
    SelectOption,
    component,
    button,
    select_menu,
    LinkButton,
)
from .groups import Group, Subgroup
from .interactable import Interactable, INTERACTION_REGISTERS

__all__ = (
    "ActionRow",
    "Button",
    "ButtonStyle",
    "ChannelTypes",
    "ChatCommandHandler",
    "Choice",
    "Choices",
    "CommandArg",
    "ComponentHandler",
    "Description",
    "Group",
    "INTERACTION_REGISTERS",
    "Interactable",
    "LinkButton",
    "MaxValue",
    "MinValue",
    "Modifier",
    "SelectMenu",
    "SelectOption",
    "Subgroup",
    "button",
    "command",
    "component",
    "message_command",
    "select_menu",
    "user_command",
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when anyone is added to or removed from a thread"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.thread import ThreadMembersUpdateEvent

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def thread_members_update_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_thread_members_update`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the thread members update event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.events.thread.ThreadMembersUpdateEvent`]
        ``on_thread_members_update`` and an ``ThreadMembersUpdateEvent``
    """  # noqa: E501
    return (
        "on_thread_members_update",
        ThreadMembersUpdateEvent.from_dict(payload.data),
    )


def export():
    return thread_members_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a message is deleted in a subscribed text channel"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.message import MessageDeleteEvent

if TYPE_CHECKING:
    from typing import Tuple

    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def on_message_delete_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
) -> Tuple[str, MessageDeleteEvent]:
    """|coro|
    Middleware for the ``on_message_delete`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the message delete event
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.events.message.MessageDeleteEvent`]
        ``on_message_delete`` and a ``MessageDeleteEvent``
    """  # noqa: E501

    return ("on_message_delete", MessageDeleteEvent.from_dict(payload.data))


def export():
    return on_message_delete_middleware



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pincer',
 'pincer.commands',
 'pincer.commands.components',
 'pincer.core',
 'pincer.middleware',
 'pincer.objects',
 'pincer.objects.app',
 'pincer.objects.events',
 'pincer.objects.guild',
 'pincer.objects.message',
 'pincer.objects.user',
 'pincer.objects.voice',
 'pincer.utils']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0']

extras_require = \
{'img': ['Pillow>=9.2.0,<10.0.0', 'types-Pillow>=9.2.1,<10.0.0'],
 'speed': ['orjson>=3.7.11,<4.0.0',
           'Brotli>=1.0.9,<2.0.0',
           'aiodns>=3.0.0,<4.0.0',
           'cchardet>=2.1.7,<3.0.0']}

setup_kwargs = {
    'name': 'pincer',
    'version': '0.16.1',
    'description': 'Discord API wrapper rebuilt from scratch.',
    'long_description': '# Pincer\n\n[![PyPI - Downloads](https://img.shields.io/badge/dynamic/json?label=downloads&query=%24.total_downloads&url=https%3A%2F%2Fapi.pepy.tech%2Fapi%2Fprojects%2FPincer)](https://pypi.org/project/Pincer)\n![PyPI](https://img.shields.io/pypi/v/Pincer)\n![PyPI - Format](https://img.shields.io/pypi/format/Pincer)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/Pincer)\n[![Scrutinizer Code Quality](https://scrutinizer-ci.com/g/Pincer-org/pincer/badges/quality-score.png?b=main)](https://scrutinizer-ci.com/g/Pincer-org/pincer/?branch=main)\n[![Build Status](https://scrutinizer-ci.com/g/Pincer-org/Pincer/badges/build.png?b=main)](https://scrutinizer-ci.com/g/Pincer-org/Pincer/build-status/main)\n[![Documentation Status](https://readthedocs.org/projects/pincer/badge/?version=latest)](https://pincer.readthedocs.io/en/latest/?badge=latest)\n[![codecov](https://codecov.io/gh/Pincer-org/Pincer/branch/main/graph/badge.svg?token=T15T34KOQW)](https://codecov.io/gh/Pincer-org/Pincer)\n![Lines of code](https://tokei.rs/b1/github/pincer-org/pincer?category=code&path=pincer)\n![Repo Size](https://img.shields.io/github/repo-size/Pincer-org/Pincer)\n![GitHub last commit](https://img.shields.io/github/last-commit/Pincer-org/Pincer)\n![GitHub commit activity](https://img.shields.io/github/commit-activity/m/Pincer-org/Pincer?label=commits)\n![GitHub](https://img.shields.io/github/license/Pincer-org/Pincer)\n![Discord](https://img.shields.io/discord/881531065859190804)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n![gitmoji](https://img.shields.io/badge/gitmoji-%20🚀%20💀-FFDD67.svg)\n\n## :pushpin: Links\n\n> <img src="../assets/svg/discord.svg" width="16px" alt="Discord Logo"> ｜Join the Discord server: https://discord.gg/pincer <br>\n> <img src="../assets/svg/pypi.svg" width="16px" alt="PyPI Logo"> ｜The PyPI package: https://pypi.org/project/Pincer <br>\n> <img src="../assets/svg/pincer.svg" width="16px" alt="Pincer Logo"> ｜Our website: https://pincer.dev <br>\n> 📝 | ReadTheDocs: https://pincer.readthedocs.io\n\n## ☄️ Installation\n\nUse the following command to install Pincer into your Python environment:\n\n```sh\npip install pincer\n```\n\nTo install our version with Aiohttp Speedup, use:\n\n```sh\npip install pincer[speed]\n```\n\n<details>\n\n<summary>\n    ⚙️ <i> Didn\'t work?</i>\n</summary>\n\nDepending on your Python installation, you might need to use one of the\nfollowing:\n\n- Python is not in PATH\n\n    ```sh\n    path/to/python.exe -m pip install pincer\n    ```\n\n- Python is in PATH but pip is not\n\n    ```sh\n    python -m pip install pincer\n    ```\n\n- Unix systems can use pip3/python3 commands\n\n    ```sh\n    pip3 install pincer\n    ```\n\n    ```sh\n    python3 -m pip install pincer\n    ```\n\n- Using multiple Python versions\n\n    ```sh\n    py -m pip install pincer\n    ```\n\n</details>\n\n**Client base class example:**\n\n```py\nfrom pincer.client import Bot\n\n# Note that both `Bot` and `Client` are valid!\nbot = Bot("YOUR_TOKEN_HERE")\nbot.run()\n```\n\n**An example on the `on_ready` event**\n\nPincer bots are required to inherit from the Client.\n\n```py\nfrom time import perf_counter\nfrom pincer import Client\n\nmarker = perf_counter()\n\n\nclass Bot(Client):\n\n    @Client.event\n    async def on_ready():\n        print(f"Logged in as {client.bot} after {perf_counter() - marker} seconds")\n\n\nclient = Bot("YOUR_TOKEN_HERE")\nclient.run()\n```\n\n### Interactions\n\nPincer makes developing application commands intuitive and fast.\n\n```py\nfrom typing import Annotation  # python 3.9+\nfrom typing_extensions import Annotation  # python 3.8\n\nfrom pincer import Client\nfrom pincer.commands import command, CommandArg, Description\nfrom pincer.objects import UserMessage, User\n\n\nclass Bot(Client):\n    @Client.event\n    async def on_ready(self) -> None:\n        ...\n\n    @command(description="Say something as the bot!")\n    async def say(self, message: str):\n        return message\n\n    @user_command\n    async def user_command(self, user: User):\n        return f"The user is {user}"\n\n    @message_command(name="Message command")\n    async def message_command(self, message: UserMessage):\n        return f"The message read \'{message.content}\'"\n\n    @command(description="Add two numbers!")\n    async def add(\n        self,\n        first: Annotation[int, Description("The first number")],\n        second: Annotation[int, Description("The second number")]\n    ):\n        return f"The addition of `{first}` and `{second}` is `{first + second}`"\n\n\n```\n\nFor more examples, you can take a look at the examples folder or check out our\nbot:\n\n> <https://github.com/Pincer-org/Pincer-bot>\n\nYou can also read the interactions guide for more information:\n> <https://docs.pincer.dev/en/latest/interactions.html>\n\n### Advanced Usage\n\n#### Enable the debug mode\n\n_If you want to see everything that is happening under the hood, either out of\ncuriosity or to get a deeper insight into the implementation of some features,\nwe provide debug logging!_\n\n```py\nimport logging\n\nlogging.basicConfig(level=logging.DEBUG)\n```\n\n#### Middleware\n\n_The middleware system was introduced in version `0.4.0-dev`. This system gives you the\nfreedom to create custom events and remove the already existing middleware created by\nthe developers. Your custom middleware directly receives the payload from\nDiscord. You can\'t do anything wrong without accessing the `override` attribute, but if\nyou do access it, the Pincer team will not provide any support for weird behavior.\nSo, in short, only use this if you know what you\'re doing. An example of using\nthe middleware system with a custom `on_ready` event can be found\n[in our docs](https://pincer.readthedocs.io/en/latest/pincer.html#pincer.client.middleware).\n._\n\n## 🏷️ License\n\n`© 2021 copyright Pincer`\n\nThis repository is licensed under the MIT License.\n\nSee LICENSE for details.\n',
    'author': 'Pincer-org',
    'author_email': 'contact@pincer.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pincer.dev',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
